# import tensorflow as tf
import os
import pickle


this_dir, this_filename = os.path.split(__file__)


class peakmodel:
    # test_model_path = os.path.join(this_dir, 'model_5_3_14_4_48.h5')
    # test_model = tf.keras.models.load_model(test_model_path)
    Model_file_t = os.path.join(this_dir, 'rfmodel_tuned.pkl')
    rf_model_t = pickle.load(open(Model_file_t, 'rb'))



import pandas as pd
import numpy as np
from itertools import groupby
from tqdm import tqdm
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn import cluster
from sklearn.ensemble import BaggingRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.manifold import TSNE
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
import scipy
from sklearn.model_selection import train_test_split
from sklearn import linear_model


def data_prep(d_input, blank_keyword, simp_summary=False, svb_thres=10,
              empty_thres=0, rt_range=[0, 30], mz_range=[0, 1200],
              sn_thres=3, score_thres=0, area_thres=5000):
    '''
    The function is used to clean the dataframe according to user setting
    blank_keyword: part of string from column that
    indicates the column is a blank sample
    svb_thres: sample vs blank thres
    empty_thres: empty cell thres in a row
    cv_thres: as all sample is in triplicate,
    calculate the CV for every triplicate sample set
    # Needs to be updated in case there is no triplicate samples
    rt_range: rt filter
    mz_range: mz filter
    sn_thres: signal/noise column thres
    score_thres: score column thres
    area_thres: count for max peak area from each row
    '''
    d_thres = d_input[d_input[d_input.columns[4:]].max(1) >= area_thres]
    d_thres = (d_thres[(d_thres['Average RT (min)'] > rt_range[0]) &
               (d_thres['Average RT (min)'] < rt_range[1])])
    d_thres = (d_thres[(d_thres['Average m/z'] > mz_range[0]) &
               (d_thres['Average m/z'] < mz_range[1])])
    d_thres = d_thres[d_thres['Average sn'] >= sn_thres]
    d_thres = d_thres[d_thres['Average score'] >= score_thres]
    d_thres.reset_index(inplace=True, drop=True)
    col_blank = []
    for key in blank_keyword:
        # Get column name if it contains blank indicating strings
        col_blank.extend([col for col in d_thres.columns if key in col])
    col_sample = [col for col in d_thres.columns if col not in col_blank]
    # Sample maximum area vs Blank average area to count for svb
    d_sample = (d_thres[d_thres[col_sample[4:]].max(axis=1) /
                d_thres[col_blank].mean(axis=1) > svb_thres][col_sample])
    d_sample.reset_index(inplace=True)
    d_sample.drop(columns=['index'], inplace=True)
    # Get a list of triplicate, every triplicate is in a sublist
    # Sample: [[a1,a2,a3],[b1,b2,b3]]
    # Note: the triplicate parsing is now only used '_'
    # which needs update in the future
    # d_transpose['dilu_vol'] = d_transpose['dilu_vol'].
    # apply(lambda x : x.replace('-','_')) in case
    # people use '-' for parsing
    trip_list = ([list(i) for j, i in groupby(d_sample.columns[4:],
                 lambda a: a.split('_')[:-1])])
    trip_list = [i for i in trip_list if len(i) >= 2]
    # filter out columns that is not in triplicate -- sample naming issue

    for triplicate in tqdm(trip_list):
        for row in d_sample[triplicate].itertuples():
            # Loop for every sets of triplicates
            if row[1:].count(0) > empty_thres:
                d_sample.loc[row.Index, triplicate] = 0
                # if more than thres, then set all three values to 0
#             elif np.mean(row[1:]) != 0:
#                 if np.std(row[1:]) / np.mean(row[1:]) > cv_thres:
#                     d_sample.loc[row.Index, triplicate] = 0
#                   need verify, not work for now
            else:
                pass
    d_sample = d_sample[~(d_sample[d_sample.columns[4:]] == 0).all(1)]
    # clean rows with all 0
    if simp_summary is True:
        simp_dict = {}
        for i, column in enumerate(trip_list):
            avg = d_sample[column].mean(1)
            cv = d_sample[column].std(1) / d_sample[column].mean(1)
            # optional display CV
            simp_dict.update({column[0][: -2]: avg, ' CV #' + str(i): cv})
        d_result = pd.DataFrame(simp_dict)
        d_result = (pd.concat([d_sample[d_sample.columns[:4]],
                    d_result], axis=1))
    elif simp_summary is False:
        d_result = d_sample.copy()
    return d_result


def ms_cluster(d_input, select_keyword, normalization='linear',
               visual=False, d_reduce=True, d_reduce_method='tsne',
               perplexity=20, cluster_method='dbscan',
               eps=0.8, min_samples=10):
    '''
    Function for direct clustering:
    normalization method: linear, zscore, log
    d_reduce: if perform the dimension reduction algorithm,
    method: only tsne is avilable now
    perplexity: parameter for tsne
    cluster_method: dbscan, later will update optic and spectrum
    eps: parameter for dbscan,
    threshold of radius that is used to count neighbours
    min_samples: general parameter for clustering,
    min neighbourhoods to be counted as a cluster
    '''
    col_select = []
    for key in select_keyword:
        col_select.extend([col for col in d_input.columns if key in col])
    d_clu = d_input[col_select]
    c_data = d_clu.values
    c_norm = []
    # Performs normalization
    np.seterr(divide='ignore', invalid='ignore')
    # silent the warning -- but divide by 0 still exist
    for row in c_data:
        if normalization == 'linear':
            c_norm.append(row/max(row))
        elif normalization == 'zscore':
            c_norm.append((row-np.mean(row))/np.std(row))
        elif normalization == 'log':
            row[row == 0] = 1
            c_norm.append(np.log10(row)/np.log10(max(row)))
        else:
            pass
    # Clean up dataframe
    c_norm = np.asarray(c_norm)
    d_norm = pd.DataFrame(c_norm)
    d_norm['index'] = d_input.index
    d_norm.set_index('index', inplace=True)
    d_norm.dropna(how='all', inplace=True)
    if d_reduce is True:
        if d_reduce_method == 'tsne':
            model = TSNE(learning_rate=100, perplexity=50, n_iter=1000)
            # Tune perplexity and n_iter
            transformed = model.fit_transform(d_norm)
            d_feature = transformed.copy()
        else:
            pass
    elif d_reduce is False:
        d_feature = d_norm.copy()
    else:
        pass
    if cluster_method == 'dbscan':
        dbscan = cluster.DBSCAN(eps=eps,
                                min_samples=min_samples).fit(d_feature)
        labels = dbscan.labels_
        unique_labels = set(dbscan.labels_)
        if visual is True:
            for i, k in enumerate(unique_labels):
                indexlist = list(np.argwhere(labels == k).reshape(1, -1)[0])
                sns.clustermap(d_norm.iloc[indexlist].values, cmap='Reds',
                               col_cluster=False, yticklabels=False,
                               xticklabels=False, figsize=(5, 5))
                plt.title(str(dbscan) + 'label=' + str(k))
                plt.show()
        else:
            pass
        d_init = d_input.copy()
        d_label = d_init.loc[d_norm.index]
        # Use the index to match back to the original datasheet
        d_label.insert(4, "label", dbscan.labels_.tolist())
    elif cluster_method == 'optics':
        optics = cluster.OPTICS(min_samples=min_samples).fit(d_feature)
        labels = optics.labels_
        unique_labels = set(optics.labels_)
        if visual is True:
            for i, k in enumerate(unique_labels):
                indexlist = list(np.argwhere(labels == k).reshape(1, -1)[0])
                sns.clustermap(d_norm.iloc[indexlist].values, cmap='Reds',
                               col_cluster=False, yticklabels=False,
                               xticklabels=False, figsize=(5, 5))
                plt.title(str(optics) + 'label=' + str(k))
                plt.show()
        else:
            pass
        d_init = d_input.copy()
        d_label = d_init.loc[d_norm.index]
        # Use the index to match back to the original datasheet
        d_label.insert(4, "label", optics.labels_.tolist())
    else:
        pass
    # Post filter -- filter out features that present in other sources
    # but not SR520 -- keep it open for now
    # If activate add one more variable:source_keyword
#     col_source = []
#     for key in source_keyword:
#         col_app = [col for col in d_thres.columns if key in col]
#         col_source += col_app
#     col_rest = [col for col in d_label.columns if col not in source][5:]
#     d_label[col_app].max(1) / d_label[col_rest].max(1)
    return d_label


def trend_calc(d_input, select_keyword, min_size=5, normalization='linear',
               method='pearsonr', visual=True):

    """This function calculates clustering
    based on the pearson correlation. It takes in a dataframe
    and a user defined value for what qualifies as a cluster.
    User can choose whether or not to have a visual
    plot of the scatter with True/False."""
    col_select = []
    for key in select_keyword:
        col_select.extend([col for col in d_input.columns if key in col])
    d_clu = d_input[col_select]
    c_data = d_clu.values
    c_norm = []
    for row in c_data:
        if normalization == 'linear':
            c_norm.append(row/max(row))
        elif normalization == 'zscore':
            c_norm.append((row-np.mean(row))/np.std(row))
        elif normalization == 'log':
            row[row == 0] = 1
            c_norm.append(np.log10(row)/np.log10(max(row)))
    c_norm = np.asarray(c_norm)
    d_norm = pd.DataFrame(c_norm)
    d_norm['index'] = d_input.index
    d_norm.set_index('index', inplace=True)
    dropna_col = d_norm.columns
    # Post treatment to fit the d_norm into original codes
    d_norm.insert(0, "RT", d_input['Average RT (min)'].tolist())
    d_norm.insert(1, "MZ", d_input['Average m/z'].tolist())
    d_norm.dropna(subset=dropna_col, inplace=True)
    d_norm.reset_index(drop=True, inplace=True)
    # Original codes
    cluster = []  # individual cluster holder
    cluster_sum = []  # total clusters
    drop_list = []  # rows that are dropped from the df
    noise = []  # list for containing noise features
    while len(d_norm) > 0:
        for row in range(len(d_norm)):
            # feature_1 = d_norm.iloc[0]
            feature_2 = d_norm.iloc[row]
            if method == 'pearsonr':
                corr, p_val = scipy.stats.pearsonr(d_norm.iloc[0, 2:],
                                                   d_norm.iloc[row, 2:])
            elif method == 'mannwhitneyu':
                corr, p_val = (scipy.stats.mannwhitneyu(d_norm.iloc[0, 2:],
                               d_norm.iloc[row, 2:], alternative=None))
            elif method == 'kruskal':
                corr, p_val = scipy.stats.kruskal(d_norm.iloc[0, 2:],
                                                  d_norm.iloc[row, 2:])
            if p_val < 0.05:
                drop_list.append(row)
                cluster += [feature_2]
            else:
                pass
        if len(cluster) <= min_size:
            noise += [cluster]
            cluster = []
        else:
            cluster_sum += [cluster]
            cluster = []
        d_norm = d_norm.drop(drop_list)
        d_norm = d_norm.reset_index(drop=True)
        drop_list = []
    append_list = []
    for i in range(len(cluster_sum)):
        for j in range(len(cluster_sum[i])):
            cluster_sum[i][j].loc['Score'] = i
            listing = np.array(cluster_sum[i][j])
            append_list.append(listing)
    cluster_df = pd.DataFrame(append_list)  # Add columns use d_clu
    append_list2 = []
    for k in range(len(noise)):
        for n in range(len(noise[k])):
            noise[k][n].loc['Score'] = -1
            listing2 = np.array(noise[k][n])
            append_list2.append(listing2)
    noise_df = pd.DataFrame(append_list2)
    final_df = pd.concat([cluster_df, noise_df])
    final_df = final_df.reset_index(drop=True)
    if visual is True:
        labels = final_df.iloc[:, -1:].values.reshape(1, -1)[0]
        unique_labels = set(labels)
        for i, k in enumerate(unique_labels):
            indexlist = list(np.argwhere(labels == k).reshape(1, -1)[0])
            sns.clustermap(final_df.iloc[indexlist, 2: -1].values, cmap='Reds',
                           col_cluster=True, yticklabels=False,
                           xticklabels=False, figsize=(5, 5))
            plt.title('trend' + 'label=' + str(k))
            plt.show()
    else:
        pass
    return final_df


def source_label(d_input, sourcelist, area_thres=5000, concat=True):
    # noise removal only based on sourcelist cols
    np.seterr(divide='ignore', invalid='ignore')
    # source labeling
    d_result = d_input.copy()
    source_col = []
    for s in sourcelist:
        source = [col for col in d_input.columns if s in col]
        source_col.append(source)
    simp_dict = {}
    for i, column in enumerate(source_col):
        avg = d_result[column].mean(1)
        cv = d_result[column].std(1) / d_result[column].mean(1)
        # optional display CV
        cv_nan = np.isnan(cv)
        cv[cv_nan] = 0.0  # replace nan with 0
        simp_dict.update({sourcelist[i]: avg, str(sourcelist[i]) + ' Cv': cv})
    d_summary = pd.DataFrame(simp_dict)
    d_summary['source'] = "NA"
    for row in d_summary.itertuples():
        sourcelabel = list(d_summary.columns[[col_index for col_index,
                           peak_avg in enumerate(row[1: -1])
                           if peak_avg >= area_thres]])
        if len(sourcelabel) != 0:
            labelstr = ','.join(sourcelabel)
            d_summary.at[row.Index, 'source'] = labelstr
    if concat is True:
        d_concat = pd.concat([d_result, d_summary], axis=1)
    elif concat is False:
        d_concat = d_result.copy()
        d_concat['source'] = d_summary['source']
    return d_concat


def source_report(d_input, source_key, mix_key, method='multiple',
                  pa_thres=10000, CV_thres=2):
    # source key needs to be the same as source_label above
    # **only take the concat dataframe from labeling function
    # prefilter & dataframe arrangement
    d_mix = (d_input[(d_input[[col for col in
             d_input.columns if 'Cv' in col]] <= CV_thres).all(1)])
    # all cv should below thres in order to be checked
    d_simp = d_mix[source_key]
    print('Threshold set to', pa_thres)
    mix_col = []
    for key in mix_key:
        mix_col.extend([col for col in d_mix.columns if 'Mix' in col])
    if len(mix_col) == 0:
        print("didn't find mixture by keyword!")

    d_st = pd.DataFrame(mix_col)

    c_name = ['sample']
    for source in source_key:
        result = []
        for col in mix_col:
            n_feature = (sum(d_mix[d_mix[col] >= pa_thres]
                         ['source'].str.contains(source)))
            cov_score = n_feature / sum(d_mix['source'].str.contains(source))
            if method == 'single':
                mix = d_mix[d_mix['source'] == source][col]
                s_simp = d_simp[d_simp['source'] == source][source]
            elif method == 'multiple':
                mix = d_mix[d_mix['source'].str.contains(source)][col]
                s_simp = (d_simp.loc[d_mix[
                          d_mix['source'].str.contains(source)].index][source])
            match_index = [i for i, j in enumerate(mix) if j >= pa_thres]
            dilu = mix.iloc[match_index] / s_simp.iloc[match_index]
            ratio_score = np.average(dilu[dilu < 1])
            result.append([n_feature, cov_score, ratio_score])
        d_st = pd.concat([d_st, pd.DataFrame(result)], axis=1)
        c_name.extend(['n_'+str(source), 'cover_s', 'ratio_s'])
    d_st.columns = c_name
    return d_st


def batch_alignment(d_ref, d_align, rt_error=0.5, mz_error=0.015):
    # Alignment of new dataset
    result = []
    for row in np.arange(len(d_ref)):
        # Find overlap using the error range
        overlap = np.where((d_align.iloc[:, 0] - rt_error <=
                           d_ref.iloc[row, 0]) & (d_ref.iloc[row, 0] <=
                           d_align.iloc[:, 0] + rt_error) &
                           (d_align.iloc[:, 1] - mz_error <=
                           d_ref.iloc[row, 1]) & (d_ref.iloc[row, 1] <=
                           d_align.iloc[:, 1] + mz_error))
        if len(overlap[0]) == 1:
            result.append([overlap[0][0], row])
        elif len(overlap[0]) > 1:
            dist = []
            for i in overlap[0]:
                # if multiple hits, use the closer feature for alignment
                (dist.append(np.sqrt(((
                 d_align.iloc[i, 0] - d_ref.iloc[row, 0])**2) +
                 ((d_align.iloc[i, 1] - d_ref.iloc[row, 1])**2))))
            result.append([overlap[0][np.argmin(dist)], row])
    # Modeling using overlapping features except noises
    align_index = [i[0] for i in result]
    ref_index = [i[1] for i in result]
    d_align = d_align.loc[align_index]
    d_ref = d_ref.iloc[ref_index]

    d_align.reset_index(drop=True, inplace=True)
    d_ref.reset_index(drop=True, inplace=True)
    d_join = d_align.iloc[:, 3:]
    # Important !! -- The indexing is important especially for other data
    d_merge = pd.concat([d_ref, d_join], axis=1)
    return d_merge


# transpose the dataframe without label column
def transpose(d_input, target_col):
    # This should be converted to a built-in function
    # within the pipeline, not having a common use case
    list_label = d_input['label']
    d_transpose = d_input.copy()
    col_model = []
    for c_dilu in target_col:
        col_model.extend([col for col in d_transpose.columns if c_dilu in col])
    d_transpose = d_transpose[col_model].T
    d_transpose.reset_index(inplace=True)
    d_transpose = d_transpose.rename(columns={'index': 'dilu_vol'})
    d_transpose['dilu_vol'] = (d_transpose['dilu_vol'].apply(
                               lambda x: x.replace('-', '_')))
    d_transpose['dilu_vol'] = (d_transpose['dilu_vol'].apply(
                               lambda x: float(x.split('_')[-2][:-2])))
    list_insert = [-2] + list(list_label)
    df_length = len(d_transpose)
    d_transpose.loc[df_length] = list_insert
    # d_transpose.sort_values(by=d_transpose.index[-1],
    # ascending=True, axis=1, inplace = True)
    return d_transpose


def feature_model(d_input, cluster_algorithm=False, test_frac=0.5,
                  model_method='linear_reg', report=False):
    """This function predicts the dilution of samples through
    two methods as that can be chosen by the user. The methods
    revolve around using all of the features or clustered features.
    It takes in two dataframes and requires the user"""
    d_transpose = d_input.copy()  # avoid pandas warning
    if cluster_algorithm is True:
        # cluster separation
        cluster_index = d_transpose.iloc[-1][1:]
        # d_concat_model = d_transpose.drop([d_transpose.index[-1]])
        # drop the index row
        xtrain, xtest, ytrain, ytest = (train_test_split(
                                        d_transpose[d_transpose.columns[1:]],
                                        d_transpose['dilu_vol'],
                                        test_size=test_frac,
                                        random_state=42))
        # reattach the cluster row back for indexing
        xtrain.reset_index(inplace=True, drop=True)
        xtest.reset_index(inplace=True, drop=True)
        ytrain.reset_index(inplace=True, drop=True)
        ytest.reset_index(inplace=True, drop=True)
        xtrain.loc[len(xtrain)] = cluster_index
        xtest.loc[len(xtest)] = cluster_index
        unique_label = list(set(cluster_index))
        # both set should have same unique_label, so no need to replicate
        list_clusterDF = []
        list_clusterDF2 = []
        for label in unique_label:
            d_cm_label = (xtrain[[i for i in xtrain.columns
                          if xtrain.iloc[-1][i] == label]])
            # index using the label row --update later to avoid warning
            d_cm_label.drop(d_cm_label.tail(1).index, inplace=True)
            list_clusterDF.append(d_cm_label)
            d_cm_label2 = (xtest[[i for i in xtest.columns
                           if xtest.iloc[-1][i] == label]])
            d_cm_label2.drop(d_cm_label2.tail(1).index, inplace=True)
            list_clusterDF2.append(d_cm_label2)
    else:
        # all features used for modelling
        cluster_index = d_transpose.iloc[-1]
        # d_concat_model = d_transpose.drop([d_transpose.index[-1]])
        xtrain, xtest, ytrain, ytest = (train_test_split(
                                        d_transpose[d_transpose.columns[1:]],
                                        d_transpose['dilu_vol'],
                                        test_size=0.5, random_state=42))
    print('data split finished!')
    if model_method == 'linear_reg':
        reg_list = []
        slope = []
        intercept = []
        scores = []
        pred = []
        if cluster_algorithm is True:
            for i in range(len(list_clusterDF)):
                model = linear_model.LinearRegression()
                reg = model.fit(list_clusterDF[i], ytrain)
                pred.append(reg.predict(list_clusterDF2[i]))
                scores.append(reg.score(list_clusterDF2[i], ytest))
                slope.append(reg.coef_)
                intercept.append(reg.intercept_)
                reg_list.append(reg)
        else:
            model = linear_model.LinearRegression()
            reg = model.fit(xtrain, ytrain)
            reg.predict(xtest)
            slope.append(reg.coef_)
            intercept.append(reg.intercept_)
    # For all below modelings, in later
    # dev should either have a reasonable default param,
    # or leave space in function for user to define params
    elif model_method == 'ensemble_rf':
        features = 5
        # this number needs to be smaller than the X otherwise will raise error
        seed = 10
        reg_list = []
        scores = []
        pred = []
        if cluster_algorithm is True:
            for i in range(len(list_clusterDF)):
                model = (RandomForestRegressor(n_estimators=1000,
                         max_features=features, random_state=seed))
                reg = model.fit(list_clusterDF[i], ytrain)
                scores.append(reg.score(list_clusterDF2[i], ytest))
                pred.append(reg.predict(list_clusterDF2[i]))
                reg_list.append(reg)
        else:
            model = (RandomForestRegressor(n_estimators=1000,
                     max_features=features, random_state=seed))
            reg = model.fit(xtrain, ytrain)
            reg.predict(xtest)
    elif model_method == 'ensemble_bagging_dt':
        num_trees = 5
        seed = 10
        reg_list = []
        pred = []
        scores = []
        if cluster_algorithm is True:
            for i in range(len(list_clusterDF)):
                model = (BaggingRegressor(
                         base_estimator=DecisionTreeClassifier(),
                         n_estimators=num_trees,
                         random_state=seed))
                reg = model.fit(list_clusterDF[i], ytrain)
                scores.append(reg.score(list_clusterDF2[i], ytest))
                pred.append(reg.predict(list_clusterDF2[i]))
                reg_list.append(reg)
        else:
            model = (BaggingRegressor(base_estimator=DecisionTreeClassifier(),
                     n_estimators=num_trees, random_state=seed))
            reg = model.fit(xtrain, ytrain)
            reg.predict(xtest)
    elif model_method == 'ensemble_bagging_svc':
        num_trees = 5
        seed = 10
        reg_list = []
        pred = []
        scores = []
        if cluster_algorithm is True:
            for i in range(len(list_clusterDF)):
                model = (BaggingRegressor(base_estimator=SVC(),
                         n_estimators=num_trees, random_state=seed))
                reg = model.fit(list_clusterDF[i], ytrain)
                scores.append(reg.score(list_clusterDF2[i], ytest))
                pred.append(reg.predict(list_clusterDF2[i]))
                reg_list.append(reg)
        else:
            model = BaggingRegressor(base_estimator=SVC(),
                                     n_estimators=num_trees, random_state=seed)
            reg = model.fit(xtrain, ytrain)
            reg.predict(xtest)
    else:
        print('please define a valid model!')
        pass
    print('modeling finished!')
    if report is True:
        print('Report starting...')
        if cluster_algorithm is True:
            reg = reg_list
            if model_method == 'linear_reg':
                print(reg[0])
                for value in range(len(slope)):
                    slope_val = slope[value]
                    intercept_val = intercept[value]
                    score_val = round(scores[value], 3)
                    ypred = pred[value]
                    print('The slope for cluster', value, 'is:', slope_val)
                    print('The intercept for cluster', value, 'is:',
                          intercept_val)
                    print('The model score for cluster', value, 'is:',
                          score_val)
                    plt.figure()
                    plt.scatter(ytest, ypred, label=value)
                    plt.xlabel('Actual dilution (mL)')
                    plt.ylabel('Predicted dilution (mL)')
                    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
                print('The model score reflects coefficient of determination.')
                print('Report finished!')
            else:
                print(reg[0])
                for value in range(len(scores)):
                    score_val = round(scores[value], 3)
                    print('The model score for cluster', value, 'is:',
                          score_val)
                    ypred = pred[value]
                    plt.figure()
                    plt.scatter(ytest, ypred, label=value)
                    plt.xlabel('Actual dilution (mL)')
                    plt.ylabel('Predicted dilution (mL)')
                    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
                print('The model score is the coefficient of determination.')
                print('Report finished!')
                pass
        else:
            if model_method == 'linear_reg':
                plt.scatter(ytest, reg.predict(xtest))
                plt.xlabel('Actual dilution (mL)')
                plt.ylabel('Predicted dilution (mL)')
                print('The model score is:', '%.3f' % reg.score(xtest, ytest))
                print('The model score is the coefficient of determination.')
                print('Report finished!')
            else:
                plt.scatter(ytest, reg.predict(xtest))
                plt.xlabel('Actual dilution (mL)')
                plt.ylabel('Predicted dilution (mL)')
                print('The model score is:', '%.3f' % reg.score(xtest, ytest))
                print('The model score is the coefficient of determination.')
                print('Report finished!')
    else:
        pass
    if cluster_algorithm is True:
        reg = reg_list
    else:
        pass
    return reg


def cluster_pred(model_list, d_merge, col_name, selected_cluster,
                 model_merge='weighted_size', confidence=0.95):
    # The function only works with cluster==True in dilu_predict
    '''The selected_cluster will be user based according to score
       report from feature_model function'''
    input_label = sorted(selected_cluster)
    unique_label = list(set(d_merge['label']))
    idx = [i for i, label in enumerate(unique_label) if label in input_label]
    # get position for input labels
    model_list = [model_list[i] for i in idx]
    # slice model list according to index
    if len(model_list) != len(input_label):
        print("model didn't match labels!")
    else:
        pass
    pred_list = []
    for i in np.arange(len(model_list)):
        (pred_list.extend(model_list[i].predict(np.asarray(d_merge[
         d_merge['label'] == input_label[i]][col_name]).reshape(1, -1))))

    def mean_confidence_interval(data, confidence):
        a = 1.0 * np.array(data)
        n = len(a)
        se = scipy.stats.sem(a)
        h = se * scipy.stats.t.ppf((1 + confidence) / 2., n-1)
        return h
    if model_merge == 'weighted_size':
        size_list = []
        for i in input_label:
            size_list.append(d_merge[d_merge['label'] == i].shape[0])
        weight_list = [i/sum(size_list) for i in size_list]
        pred = sum([p * w for p, w in zip(pred_list, weight_list)])
        # Leave the CI part out for now..
        # ci = mean_confidence_interval(pred_list, confidence=confidence)
    elif model_merge == 'average':
        pred = np.mean(pred_list)
        # ci = mean_confidence_interval(pred_list, confidence=confidence)
    return pred



import pymzml
from tqdm import tqdm
import numpy as np

from scipy.integrate import simps
import pandas as pd

import peakutils

import glob
from pathlib import Path
import scipy
import pickle
import os
# from mss import mssdata
# Modeling modules
# from tensorflow import keras
# import h5py
# from mssdata import peakmodel


def get_scans(path, ms_all=False, ms_lv=1):
    '''
    The function is used to reorganize the pymzml reading
    into a list that will have better access
    path: input mzml path
    ms_all: if you want all the ms_level be check
    ms_lv: ms_level you want to export
    '''

    # Read path using pymzml
    run = pymzml.run.Reader(path)

    scans = []
    if ms_all is False:
        for scan in run:
            if scan.ms_level == ms_lv:
                scans.append(scan)
    elif ms_all is True:
        for scan in run:
            scans.append(scan)

    return scans


# Noise removal
def noise_removal(mzml_scans, int_thres=5000):
    '''
    Remove mz&i pairs that i lower than int_thres
    from whole mzml file, looping through scans
    the output will overwrite the original mzml file
    '''
    for scan in mzml_scans:
        drop_index = np.argwhere(scan.i <= int_thres)
        scan.i = np.delete(scan.i, drop_index)
        scan.mz = np.delete(scan.mz, drop_index)

    return


# Code review
# updated to select_app, when false only select closest one, when true
# append all, use as a backdoor for now if closest algorithm messed up
def mz_locator(input_list, mz, error, select_app=True):
    '''
    Find specific mzs from given mz and error range out from a given mz array
    input list: mz list
    mz: input_mz that want to be found
    error: error range is now changed to ppm level
    '''
    target_mz = []
    target_index = []

    # ppm conversion
    error = error * 1e-6

    lower_mz = mz - error * mz
    higher_mz = mz + error * mz

    for i, mzs in enumerate(input_list):
        if mzs < lower_mz:
            continue
        elif mzs >= lower_mz:
            if mzs <= higher_mz:
                target_mz.append(mzs)
                target_index.append(i)

    if select_app is False:
        if len(target_mz) != 0:
            target_error = [abs(i - mz) for i in target_mz]
            minpos = target_error.index(min(target_error))
            t_mz = target_mz[minpos]
            t_i = target_index[minpos]
        else:
            t_mz = 0
            t_i = 'NA'
    if select_app is True:
        t_mz = target_mz
        t_i = target_index

    return t_mz, t_i


# Read model for peak assessment
this_dir, this_filename = os.path.split(__file__)
Model_file_t = os.path.join(this_dir, 'rfmodel_tuned.pkl')
rf_model_t = pickle.load(open(Model_file_t, 'rb'))
Pmodel = rf_model_t


def peak_pick(mzml_scans, input_mz, error, enable_score=True, peak_thres=0.01,
              thr=0.02, min_d=1, rt_window=1.5, peak_area_thres=1e5,
              min_scan=15, max_scan=200, max_peak=5,
              min_scan_window=20, sn_range=7):
    '''
    firstly get rt, intensity from given mz and error out of the mzml file
    Then find peak on the intensity array, represent as index --> index
    Find peak range by looping from peak index forward/backward
    until hit the peak_base --> l_range,h_range. peakspan = h_range - l_range
    Trim/correct peak range is too small or too large,
    using min_scan/max_scan,min_scan_window --> trimed l/h_range
    Integration of peak based on the given range
    using simp function --> peakarea
    '''

    # Important funciont, may need to be extracted out later
    # Data output from the chromatogram_plot function
    def ms_chromatogram_list(mzml_scans, input_mz, error):
        '''
        Generate a peak list for specific input_mz over
        whole rt period from the mzml file
        ***Most useful function!
        '''

        # Create empty list to store the data
        retention_time = []
        intensity = []
        for scan in mzml_scans:
            # print(i)
            retention_time.append(scan.scan_time[0])

            target_mz, target_index = mz_locator(scan.mz, input_mz, error)
            if len(target_index) == 0:
                intensity.append(0)
            else:
                # intensity.append(scan.i[target_index])
                # if select_app=True
                intensity.append(sum(scan.i[target_index]))

        return retention_time, intensity

    rt, intensity = ms_chromatogram_list(mzml_scans, input_mz, error)

    # Get rt_window corresponded scan number
    scan_window = int(
                      (rt_window / (rt[int(len(intensity) / 2)] -
                       rt[int(len(intensity) / 2) - 1])) / 2)
    rt_conversion_rate = rt[1] - rt[0]
    # Get peak index
    indexes = peakutils.indexes(intensity, thres=thr, min_dist=min_d)

    result_dict = {}

    # dev note: boundary detection refinement
    for index in indexes:
        h_range = index
        l_range = index
        # use relative thres, also considering S/N, 1/2 rt point?
        base_intensity = peak_thres * intensity[index]
        half_intensity = 0.5 * intensity[index]

        # Get the higher and lower boundary
        while intensity[h_range] >= base_intensity:
            h_range += 1
            if intensity[h_range] < half_intensity:  # potentially record this
                if h_range - index > 4:  # fit r2 score,
                    # keep record
                    # https://stackoverflow.com/questions/55649356/
                    # how-can-i-detect-if-trend-is-increasing-or-
                    # decreasing-in-time-series as alternative
                    x = np.linspace(h_range - 2, h_range, 3)
                    y = intensity[h_range - 2: h_range + 1]
                    (slope, intercept, r_value,
                     p_value, std_err) = scipy.stats.linregress(x, y)
                    # print(rt[h_range],r_value)
                    if abs(r_value) < 0.6:
                        break
                    elif h_range > len(intensity) - 2:
                        break
        # Dev part 2, low priority since general peak shapes
        while intensity[l_range] >= base_intensity:
            l_range -= 1
            if intensity[l_range] < half_intensity:
                pass  # backdoor for recording 1/2 rt point
        # Output a range from the peak list

        peak_range = []
        if h_range - l_range >= min_scan:
            if rt[h_range] - rt[l_range] <= rt_window:
                peak_range = intensity[l_range:h_range]
            else:
                l_range = index - scan_window
                h_range = index + scan_window
                peak_range = intensity[l_range:h_range]
                # print(index + scan_window)

        # Calculate for S/N
        signal = intensity[index]
        neighbour_blank = (intensity[
                           l_range - sn_range: l_range] +
                           intensity[h_range + 1: h_range +
                           sn_range + 1])
        noise = max(neighbour_blank)
        if noise != 0:
            sn = round(signal / noise, 3)
        elif noise == 0:
            sn = 0

        # Calculate height/width, consider log10 transform
        height = signal
        width = rt[h_range] - rt[l_range]
        hw_ratio = round(height / width, 0)

        # Additional global parameters
        # 1/2 peak range
        h_loc = index
        l_loc = index
        while intensity[h_loc] > half_intensity:
            h_loc += 1
        while intensity[l_loc] > half_intensity:
            l_loc -= 1
        # calculate for slope -- interpolation included-- pay attention!
        h_half = h_loc + \
            (half_intensity - intensity[h_loc]) / \
            (intensity[h_loc - 1] - intensity[h_loc])
        l_half = l_loc + \
            (half_intensity - intensity[l_loc]) / \
            (intensity[l_loc + 1] - intensity[l_loc])
        # when transfer back use rt[index] instead
        mb = (height - half_intensity) / \
            ((h_half - index) * rt_conversion_rate)
        ma = (height - half_intensity) / \
            ((index - l_half) * rt_conversion_rate)

        # Intergration based on the simps function
        if len(peak_range) >= min_scan:
            integration_result = simps(peak_range)
            if integration_result >= peak_area_thres:
                # Calculate Area/background ratio, i.e, peak area vs
                # rectangular area as whole(if =1 then peak is a pleateu)
                background_area = (h_range - l_range) * height
                ab_ratio = round(integration_result / background_area, 3)

                # Awaiting to be added:
                # model prediction as the assessment score column
                # score = model.fit(X_para from all above)
                if enable_score is True:
                    w = rt[h_range] - rt[l_range]
                    t_r = (h_half - l_half) * rt_conversion_rate
                    l_width = rt[index] - rt[l_range]
                    r_width = rt[h_range] - rt[index]
                    assym = r_width / l_width
                    var = (w ** 2 / (1.764 * ((r_width / l_width)
                           ** 2) - 11.15 * (r_width / l_width) + 28))
                    x_peak = [w, t_r, l_width, r_width, assym,
                              integration_result, sn, hw_ratio, ab_ratio,
                              height, ma, mb, ma + mb, mb / ma, var]
                    x_input = np.asarray(x_peak)
                    # score = np.argmax(Pmodel.predict(x_input.reshape(1,-1)))
                    # for tensorflow
                    score = int(Pmodel.predict(x_input.reshape(1, -1)))
                elif enable_score is False:
                    score = 1
                # final result append score

                # appending to result
                if len(result_dict) == 0:
                    (result_dict.update(
                     {index: [l_range, h_range,
                      integration_result, sn, score]}))
                # Compare with previous item
                elif integration_result != list(result_dict.values())[-1][2]:
                    s_window = abs(index - list(result_dict.keys())[-1])
                    if s_window > min_scan_window:
                        (result_dict.update(
                         {index: [l_range, h_range, integration_result,
                          sn, score]}))

        # Filtering:
        # 1. delete results that l_range/h_range within 5 scans
        # 3. If still >5 then select top 5 results
        # list(result_dict.values())[-1]

    # Noise filter
    if len(result_dict) > max_peak:
        result_dict = {}

    return result_dict


# Code review
def peak_list(mzml_scans, err_ppm=20, enable_score=True, mz_c_thres=5,
              peak_base=0.005, thr=0.02, min_d=1, rt_window=1.5,
              peak_area_thres=1e5, min_scan=7, scan_thres=7):
    '''
    Generate a dataframe by looping throughout the
    whole mz space from a given mzml file
    ref to peak_picking function
    Q to solve: how to correctly select mz slice?? see mz_locator
    '''

    # Get m/z range -- updated 0416
    print('Generating mz list...')

    # Function to filter out empty mz slots to speed up the process
    def mz_gen(mzml_scans, err_ppm, mz_c_thres):
        pmz = []
        for scan in mzml_scans:
            pmz.append(scan.mz)
        pmz = np.hstack(pmz).squeeze()

        # Function to generate a reference mz list using a defined step
        # according to user setting
        def mz_list_gen(minmz, maxmz, error_ppm):
            error = error_ppm * 1e-6
            mz_list = [minmz]
            mz = minmz
            while mz <= maxmz:
                mz = mz + 2 * error * mz
                mz_list.append(mz)
            return mz_list

        mz_list = mz_list_gen(pmz.min(), pmz.max(), err_ppm)

        final_mz = []
        for m in mz_list:
            lm = m - err_ppm * 1e-6 * m
            hm = m + err_ppm * 1e-6 * m
            if len(pmz[(pmz <= hm) & (pmz >= lm)]) >= mz_c_thres:
                final_mz.append(m)

        return final_mz

    mzlist = mz_gen(mzml_scans, err_ppm, mz_c_thres)  # New list for looping
    print('Finding peaks...')

    result_dict = {}
    rt = []
    for scan in mzml_scans:
        rt.append(scan.scan_time[0])

    for mz in tqdm(mzlist):
        try:
            peak_dict = peak_pick(mzml_scans, mz, err_ppm, enable_score)
        except Exception:
            pass

        if len(peak_dict) != 0:
            if len(result_dict) == 0:
                for index in peak_dict:
                    result_dict.update({'m/z': [mz],
                                        'rt': [rt[index]],
                                        'sn': [peak_dict[index][3]],
                                        'score': [peak_dict[index][4]],
                                        'peak area': [peak_dict[index][2]]})
            else:
                for index in peak_dict:
                    result_dict['m/z'].append(mz)
                    result_dict['rt'].append(rt[index])
                    result_dict['sn'].append(peak_dict[index][3])
                    result_dict['score'].append(peak_dict[index][4])
                    result_dict['peak area'].append(peak_dict[index][2])
    # print(result_dict)
    print('Peak processing finished!')
    d_result = pd.DataFrame(result_dict)
    d_result['rt'] = round(d_result['rt'], 2)
    d_result['m/z'] = round(d_result['m/z'], 4)
    print('Dataframe created!')

    return d_result


# Only work on MS1 scans, needs update on the MS2 included scans
def batch_scans(path, remove_noise=True, thres_noise=1000):
    all_files = glob.glob(path + "/*.mzML")
    scans = []
    file_list = []
    for file in tqdm(all_files):
        scan = get_scans(file)
        if remove_noise is True:
            noise_removal(scan, thres_noise)
        scans.append(scan)
        file_list.append(Path(file).name)
    print(file_list)
    print('Batch read finished!')

    return scans, file_list


def batch_peak(batch_input, source_list, mz, error):
    rt_max = []
    rt_start = []
    rt_end = []
    peak_area = []
    source = []
    for i, scans in enumerate(batch_input):
        rt = []
        result_dict = peak_pick(scans, mz, error)
        for scan in scans:
            rt.append(scan.scan_time[0])
        for index in result_dict:
            rt_max.append(round(rt[index], 2))
            rt_start.append(round(rt[list(result_dict.values())[0][0]], 2))
            rt_end.append(round(rt[list(result_dict.values())[0][1]], 2))
            peak_area.append(round(result_dict[index][2], 4))
            source.append(source_list[i])
    result_dict = {'rt_max': rt_max,
                   'rt_start': rt_start,
                   'rt_end': rt_end,
                   'peak_area': peak_area,
                   'source': source
                   }
    d_result = pd.DataFrame(result_dict)

    return d_result



import unittest
import os
import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
import re
from scipy.integrate import simps
import peakutils
import webbrowser
import pyperclip
import mss
from mss import visreader


class test_dm(unittest.TestCase):
    def test_mz_locator(self):
        test_array = [0, 10, 20, 50, 100]
        result = visreader.mz_locator(test_array, 49.9999, 20)
        assert result[0][0] == 50, 'Wrong mz found'
        assert result[1][0] == 3, 'Wrong index found'
        return

    def test_formula_mass(self):
        test_formula = 'C18H22N2O2'
        result = visreader.formula_mass(test_formula)
        assert result == 299.1764085799, 'Wrong output'
        return



import os
import unittest

import pandas as pd
import numpy as np
import mss
from mss import alignment

data_path = os.path.dirname(os.path.join(mss.__path__[0]))
batchpath = data_path + '/example_data/peakdata/mssoutput/'


class test_alignment(unittest.TestCase):
    def test_stack(self):
        """Tests stack functionlity"""
        a, b = alignment.stack(batchpath)
        assert a > 0, "no files in the folder"
        assert len(b) > 0, "output size is wrong"
        assert type(b) == pd.core.frame.DataFrame, "output type is wrong"
        assert b[b.columns[0]].dtypes == np.float32, "stacking seems off"
        assert len(b.columns) > 0, 'no columns'

    def test_realignment(self):
        """Tests realignment functionality"""
        batch_name = 'test1'
        rt_error = 0.05
        MZ_error = 0.015
        file_type = '.csv'
        b = alignment.realignment(batchpath, batch_name, file_type,
                                  rt_error, MZ_error)
        assert len(b) > 0, "testing realingment length is off"
        assert type(b) == pd.core.frame.DataFrame, "output type is wrong"
        assert b[b.columns[0]].dtypes == np.float32, "wrong value type"
        assert b.columns[0] == 'Average m/z', 'wrong column'
        assert b.columns[1] == 'Average RT (min)', 'wrong column'
        assert b.columns[2] == 'Average sn', 'wrong column'
        assert b.columns[3] == 'Average score', 'wrong column'



import pandas as pd
import numpy as np
from tqdm import tqdm
import os
"""This file contains the functions needed to correct the peak positions
of different samples accumulated in one day's worth of analysis."""


def stack(batch_path):
    """This function compiles all samples files into one dataframe
    for analysis. It takes in files that are of the .txt type."""
    all_samples = []
    # Calculating the number of files in path
    num_files = len([f for f in os.listdir(batch_path)if
                    os.path.isfile(os.path.join(batch_path, f))])
    print("Reading in files...")
    # Reading in files into a list
    for files in range(num_files):
        sample = os.listdir(batch_path)[files]
        sample_df = pd.read_csv(batch_path + sample,
                                dtype=np.float32)
        sample_df = sample_df.drop([sample_df.columns[0]], axis=1)
        all_samples += [sample_df]
    # Combining all the dataframes into one
    total_samples = pd.concat(all_samples)
    all_samples.clear()
    # Cleaning up the data before processing
    total_samples.loc[total_samples.sn == 0, 'sn'] = float('inf')
    total_samples.loc[total_samples.score == 3.0, 'score'] = 0.1
    total_samples.loc[total_samples.score == 2.0, 'score'] = 0.6
    print("Process completed!")
    return num_files, total_samples


def realignment(batch_path, batch_name, file_type, rt_error, MZ_error):
    """This function works by using one .txt file as a reference in which
    other files realigned to in terms of precursor and RT. """
    RT_error = rt_error  # units of minutes, can be adjusted
    alignment_df = pd.DataFrame()
    mz_error = MZ_error
    standard_sample = os.listdir(batch_path)[0]  # first sample
    # reads .txt file into a dataframe
    standard_df = pd.read_csv(batch_path + standard_sample, usecols=['rt',
                              'm/z', 'sn', 'score', 'peak area'],
                              dtype=np.float32)
    # Creating the reference df based off 1st sample
    alignment_df['m/z'] = standard_df.iloc[:, 0]
    alignment_df['rt'] = standard_df.iloc[:, 1]
    alignment_df['sn'] = standard_df.iloc[:, 2]
    alignment_df['score'] = standard_df.iloc[:, 3]
    alignment_df['Sum RT (min)'] = 0.0
    alignment_df['Sum Precursor m/z'] = 0.0
    alignment_df['Sum sn'] = 0.0
    alignment_df['Sum score'] = 0.0
    alignment_df['Sum RT (min)'] = alignment_df[
                                   'Sum RT (min)'].astype('float32')
    alignment_df['Sum Precursor m/z'] = alignment_df[
                                        'Sum Precursor m/z'].astype('float32')
    alignment_df['Sum sn'] = alignment_df['Sum sn'].astype('float32')
    alignment_df['Sum score'] = alignment_df['Sum score'].astype('float32')
    alignment_df.loc[alignment_df.sn == 0, 'sn'] = float('inf')
    alignment_df.loc[alignment_df.score == 3.0, 'score'] = 0.1
    alignment_df.loc[alignment_df.score == 2.0, 'score'] = 0.6
    # col_index to track where to add sample columns
    col_index = 8
    num_files, total_samp = stack(batch_path)
    for files in range(num_files):
        sample = os.listdir(batch_path)[files]  # Creates file columns
        sample = sample[:-4]  # removes extension part in sample name
        alignment_df[sample] = 0.0
        alignment_df[sample] = alignment_df[sample].astype('float32')
    print("Initial reference built")
    print("Alignment beginning..")
    for row in tqdm(range(len(total_samp))):
        if alignment_df.isnull().values.any() is True:
            alignment_df = alignment_df.fillna(0)
        else:
            pass
        row_max = len(alignment_df)
        if row < range(len(total_samp))[-1]:
            # Loops until a change in sample occurs by index
            if total_samp.index[row] < total_samp.index[row + 1]:
                # checks to see if row exists in reference or not
                overlap = np.where(((alignment_df.iloc[:, 1] - RT_error)
                                   <= total_samp.iloc[row, 1]) &
                                   (total_samp.iloc[row, 1] <=
                                   (alignment_df.iloc[:, 1] + RT_error))
                                   & ((alignment_df.iloc[:, 0] - mz_error)
                                      <= total_samp.iloc[row, 0])
                                   & (total_samp.iloc[row, 0] <=
                                       (alignment_df.iloc[:, 0] + mz_error)))
                if len(overlap[0]) > 0:
                    overlap = [overlap][0][0][0]
                    # checks for any duplicates
                    if (alignment_df.iloc[overlap, col_index]) > 0:
                        alignment_df.at[overlap, alignment_df.columns[
                                        col_index]] += total_samp.iloc[row, 4]
                        # Averages the mean if there is a repeat intensity
                        alignment_df.at[overlap, alignment_df.columns[
                                         col_index]] /= 2
                    else:
                        # sets intensity value where realignment was identified
                        alignment_df.at[overlap, alignment_df.columns[
                                             col_index]] = total_samp.iloc[
                                                           row, 4]
                        # adds rt, sn, m/z, score to aligned point
                        alignment_df.at[overlap,
                                        'Sum RT (min)'] += (total_samp.
                                                            iloc[row, 1])
                        alignment_df.at[overlap,
                                        'Sum Precursor m/z'] += (total_samp.
                                                                 iloc[row, 0])
                        alignment_df.at[overlap, 'Sum sn'] += (total_samp.
                                                               iloc[row, 2])
                        alignment_df.at[overlap, 'Sum score'] += (total_samp.
                                                                  iloc[row, 3])
                else:
                    # appends as a new reference if nothing is found
                    alignment_df.loc[row_max] = total_samp.iloc[row, 0:4]
                    alignment_df.at[alignment_df.index[
                                   -1], alignment_df.columns[
                                        col_index]] = total_samp.iloc[row, 4]
                    alignment_df.at[alignment_df.index[
                                   -1], 'Sum RT (min)'] = total_samp.iloc[
                                                          row, 1]
                    alignment_df.at[alignment_df.index[
                                   -1], 'Sum Precursor m/z'] = total_samp.iloc[
                                                               row, 0]
                    alignment_df.at[alignment_df.index[
                                   -1], 'Sum sn'] = total_samp.iloc[row, 2]
                    alignment_df.at[alignment_df.index[
                                   -1], 'Sum score'] = total_samp.iloc[row, 3]
            else:
                # if index changes, moves to next sample column
                col_index += 1
                overlap = np.where(((alignment_df.iloc[:, 1] - RT_error)
                                   <= total_samp.iloc[row, 1]) &
                                   (total_samp.iloc[row, 1] <=
                                   (alignment_df.iloc[:, 1] + RT_error))
                                   & ((alignment_df.iloc[:, 0] - mz_error)
                                      <= total_samp.iloc[row, 0])
                                   & (total_samp.iloc[row, 0] <=
                                       (alignment_df.iloc[:, 0] + mz_error)))
                if len(overlap[0]) > 0:
                    overlap = [overlap][0][0][0]
                    if (alignment_df.iloc[overlap, col_index]) > 0:
                        alignment_df.at[overlap, alignment_df.columns[
                                         col_index]] += total_samp.iloc[row, 4]
                        # Averages the mean if there is a repeat intensity
                        alignment_df.at[overlap, alignment_df.columns[
                                         col_index]] /= 2
                    else:
                        alignment_df.at[overlap, alignment_df.columns[
                                             col_index]] = total_samp.iloc[
                                                           row, 4]
                        # adds rt, m/z, score, sn to aligned point
                        alignment_df.at[overlap,
                                        'Sum RT (min)'] += (total_samp.
                                                            iloc[row, 1])
                        alignment_df.at[overlap,
                                        'Sum Precursor m/z'] += (total_samp.
                                                                 iloc[row, 0])
                        alignment_df.at[overlap, 'Sum sn'] += (total_samp.
                                                               iloc[row, 2])
                        alignment_df.at[overlap, 'Sum score'] += (total_samp.
                                                                  iloc[row, 3])
                else:
                    # appends as a new reference if nothing is found
                    alignment_df.loc[row_max] = total_samp.iloc[row, 0:4]
                    alignment_df.at[alignment_df.index[
                                   -1], alignment_df.columns[
                                        col_index]] = total_samp.iloc[row, 4]
                    alignment_df.at[alignment_df.index[
                                   -1], 'Sum RT (min)'] = total_samp.iloc[
                                                          row, 1]
                    alignment_df.at[alignment_df.index[
                                   -1], 'Sum Precursor m/z'] = total_samp.iloc[
                                                               row, 0]
                    alignment_df.at[alignment_df.index[
                                   -1], 'Sum sn'] = total_samp.iloc[row, 2]
                    alignment_df.at[alignment_df.index[
                                   -1], 'Sum score'] = total_samp.iloc[row, 3]
        else:  # for last row of dataframe
            overlap = np.where(((alignment_df.iloc[:, 1] - RT_error)
                               <= total_samp.iloc[row, 1]) &
                               (total_samp.iloc[row, 1] <=
                                (alignment_df.iloc[:, 1] + RT_error))
                               & ((alignment_df.iloc[:, 0] - mz_error)
                                   <= total_samp.iloc[row, 0])
                               & (total_samp.iloc[row, 0]
                                   <= (alignment_df.iloc[:, 0] + mz_error)))
            if len(overlap[0]) > 0:
                overlap = [overlap][0][0][0]
                if (alignment_df.iloc[overlap, col_index]) > 0:
                    alignment_df.at[overlap, alignment_df.columns[
                                         col_index]] += total_samp.iloc[row, 4]
                    # Averages the mean if there is a repeat intensity
                    alignment_df.at[overlap, alignment_df.columns[
                                         col_index]] /= 2
                else:
                    # sets intensity value where realignment was identified
                    alignment_df.at[overlap, alignment_df.columns[
                                         col_index]] = total_samp.iloc[row, 4]
                    # adds rt, m/z, score, sn to aligned point
                    alignment_df.at[overlap, 'Sum RT (min)'] += (total_samp.
                                                                 iloc[row, 1])
                    alignment_df.at[overlap,
                                    'Sum Precursor m/z'] += (total_samp.
                                                             iloc[row, 0])
                    alignment_df.at[overlap, 'Sum sn'] += total_samp.iloc[
                                                           row, 2]
                    alignment_df.at[overlap, 'Sum score'] += total_samp.iloc[
                                                              row, 3]
            else:
                # appends as a new reference if nothing is found
                alignment_df.loc[row_max] = total_samp.iloc[row, 0:4]
                alignment_df.at[alignment_df.index[
                               -1], alignment_df.columns[
                                    col_index]] = total_samp.iloc[row, 4]
                alignment_df.at[alignment_df.index[
                               -1], 'Sum RT (min)'] = total_samp.iloc[row, 1]
                alignment_df.at[alignment_df.index[
                               -1], 'Sum Precursor m/z'] = total_samp.iloc[
                                                           row, 0]
                alignment_df.at[alignment_df.index[
                               -1], 'Sum sn'] = total_samp.iloc[row, 2]
                alignment_df.at[alignment_df.index[
                               -1], 'Sum score'] = total_samp.iloc[row, 3]
    alignment_df.rename(columns={'rt': 'Average RT (min)'}, inplace=True)
    alignment_df.rename(columns={'m/z': 'Average m/z'}, inplace=True)
    alignment_df.rename(columns={'sn': 'Average sn'}, inplace=True)
    alignment_df.rename(columns={'score': 'Average score'}, inplace=True)
    # Replace all NaN elements with 0
    alignment_df = alignment_df.fillna(0)
    # final dataframe sorted by m/z
    alignment_df = alignment_df.sort_values(by='Average m/z',
                                            ignore_index=True)
    # empty list to hold invalid peaks
    invalid = []
    for rows in range(len(alignment_df)):
        # Calculating the averages after the iterations
        # requires count of nonzero count to calculate the mean properly
        count = np.count_nonzero(alignment_df.iloc[rows, 8:])
        if count > 0:
            alignment_df.at[rows, 'Average RT (min)'] = (alignment_df.loc[rows,
                                                         'Sum RT (min)']/count)
            alignment_df.at[rows, 'Average m/z'] = (alignment_df.loc[rows,
                                                    'Sum Precursor m/z']/count)
            alignment_df.at[rows, 'Average sn'] = (alignment_df.loc[rows,
                                                   'Sum sn']/count)
            alignment_df.at[rows, 'Average score'] = (alignment_df.loc[rows,
                                                      'Sum score']/count)
        else:
            invalid.append(rows)
            print("Identified invalid peak(s)")  # For detection error
    if len(invalid) > 0:
        for i in range(len(invalid)):
            alignment_df = alignment_df.drop(alignment_df.index[invalid[i]])
        alignment_df = alignment_df.reset_index(drop=True)
        print("Removed invalid peak(s) from reference")
    else:
        pass
    # Drop columns to collect sums for averaging
    alignment_df = alignment_df.drop(columns=[
                   'Sum RT (min)', 'Sum Precursor m/z',
                   'Sum sn', 'Sum score'])
    # rounds numbers in the first two columns
    alignment_df = alignment_df.round({'Average RT (min)': 3,
                                       'Average m/z': 5})
    print("Alignment done!")
    # converts file for saving
    alignment_df.to_csv(batch_path + batch_name +
                        file_type, header=True, index=False)
    print("File saved")
    return alignment_df



import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
import re
from scipy.integrate import simps
import peakutils
import webbrowser
import pyperclip


# TIC plot
def tic_plot(mzml_scans, interactive=True):
    '''
    Static tic plot function
    '''
    time = []
    TIC = []
    for scan in mzml_scans:
        time.append(scan.scan_time[0])
        TIC.append(scan.TIC)

    if interactive is True:
        fig = go.Figure([go.Scatter(x=time, y=TIC,
                        hovertemplate='Int: %{y}' + '<br>RT: %{x}minute<br>')])

        fig.update_layout(
            template='simple_white',
            width=1000,
            height=600,
            xaxis=dict(title='Retention Time (min)',
                       rangeslider=dict(visible=True)),
            yaxis=dict(
                showexponent='all',
                exponentformat='e',
                title='Intensity',
            ))

        fig.show()

    elif interactive is False:
        plt.figure(figsize=(10, 6))
        plt.plot(time, TIC)
        plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        plt.xlabel('RT (min)')
        plt.ylabel('TIC')
        plt.title('TIC spectrum')
        plt.show()

    return


def ms_plot(mzml_scans, time, interactive=False, search=False, source='MoNA'):
    '''
    Interactive spectrum plot with nearest retention time from the given scan
    mzml_scans: mzfile
    time: selected time for the scan
    '''
    for scan in mzml_scans:
        if scan.scan_time[0] >= time:
            mz = scan.mz
            ints = scan.i
            rt = scan.scan_time[0]
            break

    if interactive is True:
        plt.clf()
        fig = go.Figure([go.Bar(x=mz, y=ints, marker_color='red', width=0.5,
                         hovertemplate='Int: %{y}' + '<br>m/z: %{x}<br>')])
        fig.update_traces(marker_color='rgb(158,202,225)',
                          marker_line_color='rgb(0,0,0)',
                          marker_line_width=0.5, opacity=1)
        fig.update_layout(
                title_text=str(round(rt, 3)) +
                ' min MS1 spectrum, input ' + str(time) + ' min',
                template='simple_white',
                width=1000,
                height=600,
                xaxis={'title': 'm/z ratio'},
                yaxis=dict(
                    showexponent='all',
                    exponentformat='e',
                    title='Intensity'))
        fig.show()

    elif interactive is False:
        plt.figure(figsize=(10, 5))
        plt.bar(mz, ints, width=1.0)
        plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        plt.xlabel('m/z')
        plt.ylabel('Intensity')
        plt.title('MS1 spectrum')

    if search is True:
        for i in range(len(mz)):
            if i == 0:
                list_string = str(round(mz[i], 4)) + ' '
                + str(round(ints[i], 1)) + '\r'
            else:
                list_string += str(round(mz[i], 4)) + ' '
                + str(round(ints[i], 1)) + '\r'
        pyperclip.copy(list_string)
        if source == 'MoNA':
            webbrowser.open("https://mona.fiehnlab.ucdavis.edu/spectra/search")
        elif source == 'metfrag':
            webbrowser.open("https://msbi.ipb-halle.de/MetFragBeta/")

    return


def frag_plot(mzml_scans, precursor, error=20, scan_index=0,
              noise_thr=50, interactive=False, search=False, source='MoNA'):
    '''
    Interactive spectrum plot with nearest retention time from the given scan
    mzml_scans: mzfile
    time: selected time for the scan
    '''
    p_range_l = precursor * (1 - error * 1e-6)
    p_range_h = precursor * (1 + error * 1e-6)
    frag_scan = []
    for scan in mzml_scans:
        if scan.ms_level == 2:
            precursor = scan.selected_precursors[0]['mz']
            p_intensity = scan.selected_precursors[0]['i']
            if precursor < p_range_h and precursor > p_range_l:
                frag_scan.append([precursor, p_intensity, scan])
    frag_scan.sort(key=lambda x: x[1], reverse=True)
    if len(frag_scan) != 0:
        print('Now showing index', scan_index, 'of',
              str(len(frag_scan)), 'total found scans')
        plot_scan = frag_scan[scan_index][2]
        drop_index = np.argwhere(plot_scan.i <= noise_thr)
        plot_scan.i = np.delete(plot_scan.i, drop_index)
        plot_scan.mz = np.delete(plot_scan.mz, drop_index)
        mz = plot_scan.mz
        ints = plot_scan.i
        rt = plot_scan.scan_time[0]
        print('Precursor:', round(plot_scan.selected_precursors[0]['mz'], 4),
              'precursor intensity:',
              round(plot_scan.selected_precursors[0]['i'], 1))
        print('Scan time:', round(plot_scan.scan_time[0], 2), 'minute')

        if interactive is True:
            plt.clf()
            fig = go.Figure([go.Bar(x=mz, y=ints,
                                    marker_color='red', width=0.5,
                                    hovertemplate='Int: %{y}' +
                                    '<br>m/z: %{x}<br>')])
            fig.update_traces(marker_color='rgb(158,202,225)',
                              marker_line_color='rgb(0,0,0)',
                              marker_line_width=0.5, opacity=1)
            fig.update_layout(
                    title_text=str(round(rt, 3)) +
                    ' min MS1 spectrum, input ' + str(rt) + ' min',
                    template='simple_white',
                    width=1000,
                    height=600,
                    xaxis={'title': 'm/z ratio'},
                    yaxis=dict(
                        showexponent='all',
                        exponentformat='e',
                        title='Intensity'))
            fig.show()

        elif interactive is False:
            plt.figure(figsize=(10, 5))
            plt.bar(mz, ints, width=1.0)
            plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
            plt.xlabel('m/z')
            plt.ylabel('Intensity')
            plt.title('MS1 spectrum')
            plt.xlim(0,)

        if search is True:
            for i in range(len(mz)):
                if i == 0:
                    list_string = str(round(mz[i], 4)) + ' '
                    + str(round(ints[i], 1)) + '\r'
                else:
                    list_string += str(round(mz[i], 4)) + ' '
                    + str(round(ints[i], 1)) + '\r'
            pyperclip.copy(list_string)
            if source == 'MoNA':
                (webbrowser.open
                 ("https://mona.fiehnlab.ucdavis.edu/spectra/search"))
            elif source == 'metfrag':
                webbrowser.open("https://msbi.ipb-halle.de/MetFragBeta/")

    else:
        print('No MS2 spectrum found!')

    return


def mz_locator(input_list, mz, error, select_app=True):
    # updated to select_app, when false only select closest one,
    # when true append all, use as a backdoor
    # for now if closest algorithm messed up
    '''
    Find specific mzs from given mz and error range out from a given mz array
    input list: mz list
    mz: input_mz that want to be found
    error: error range is now changed to ppm level
    '''
    target_mz = []
    target_index = []

    # ppm conversion
    error = error * 1e-6

    lower_mz = mz - error * mz
    higher_mz = mz + error * mz

    for i, mzs in enumerate(input_list):
        if mzs < lower_mz:
            continue
        elif mzs >= lower_mz:
            if mzs <= higher_mz:
                target_mz.append(mzs)
                target_index.append(i)

    if select_app is False:
        if len(target_mz) != 0:
            target_error = [abs(i - mz) for i in target_mz]
            minpos = target_error.index(min(target_error))
            t_mz = target_mz[minpos]
            t_i = target_index[minpos]
        else:
            t_mz = 0
            t_i = 'NA'
    if select_app is True:
        t_mz = target_mz
        t_i = target_index

    return t_mz, t_i


def formula_mass(input_formula, mode='pos'):
    '''
    sudo code:
    convert input string into a list with element:number structure
    convert all the element into upper case
    match the string list into a given list of element weight
    add adduct/delete H according to mode -- also have neutral mode
    '''
    # Define a list
    elist = {'C': 12,
             'H': 1.00782,
             'N': 14.0031,
             'O': 15.9949,
             'S': 31.9721,
             'P': 30.973763,
             'e': 0.0005485799}

    mol_weight = 0
    parsed_formula = re.findall(r'([A-Z][a-z]*)(\d*)', input_formula)
    for element_count in parsed_formula:
        element = element_count[0]
        count = element_count[1]
        if count == '':
            count = 1

        mol_weight += elist[element]*float(count)

    if mode == 'pos':
        mol_weight += elist['e'] + elist['H']
    elif mode == 'neg':
        mol_weight -= elist['e'] + elist['H']
    else:
        pass

    return mol_weight


def ms_chromatogram(mzml_scans, input_value, error,
                    smooth=False, mode='pos', interactive=True,
                    search=False, source='pubchem'):
    '''
    Interactive chromatogram for selected m/z
    search is now only available on chemspider and pubchem
    '''
    if type(input_value) == float:
        input_mz = input_value
    elif type(input_value) == int:
        input_mz = input_value
    elif type(input_value) == str:
        input_mz = formula_mass(input_value, mode)
    else:
        print('Cant recognize input type!')

    print(round(input_mz, 3))

    retention_time = []
    intensity = []
    for scan in mzml_scans:
        # print(i)
        retention_time.append(scan.scan_time[0])

        target_mz, target_index = mz_locator(scan.mz, input_mz, error)
        if target_index == 'NA':
            intensity.append(0)
        else:
            intensity.append(sum(scan.i[target_index]))

    def peak_smooth(input_list, baseline=500):
        for i, int_ in enumerate(input_list):
            if i > 1 and i < len(input_list)-3:
                if int_ > baseline:
                    for index in np.arange(i+1, i+3):
                        if input_list[index] == 0:
                            input_list[index] = (input_list[index-1] +
                                                 input_list[index+1])/2
                        else:
                            continue

    if smooth is True:
        peak_smooth(intensity)

    if interactive is True:
        fig = go.Figure([go.Scatter(x=retention_time, y=intensity,
                        hovertemplate='Int: %{y}' + '<br>RT: %{x}minute<br>')])

        fig.update_layout(
            title_text=str(round(input_mz, 2)) +
            ' chromatogram, error ' + str(error),
            template='simple_white',
            width=1000,
            height=600,
            xaxis={'title': 'Retention Time (min)'},
            yaxis=dict(
                showexponent='all',
                exponentformat='e',
                title='Intensity'))

        fig.show()
    elif interactive is False:
        plt.figure(figsize=(20, 10))
        plt.plot(retention_time, intensity)
        plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        plt.xlabel('Retention Time(min)')
        plt.ylabel('Intensity')
        plt.title('MS1 spectrum')
        plt.xlim(0, retention_time[-1])
        plt.ylim(0, )
        plt.show()

    if search is False:
        pass
    if search is True:
        if type(input_value) == str:
            if source == 'chemspider':
                webbrowser.open("http://www.chemspider.com/Search.aspx?q="
                                + input_value)
            elif source == 'pubchem':
                webbrowser.open("https://pubchem.ncbi.nlm.nih.gov/#query="
                                + input_value)
        else:
            print('Please enter formula for search!')

    return


# Deal with roi, check peakonly
def peak_pick(mzml_scans, input_mz, error, peak_base=5000,
              thr=0.02, min_d=1, rt_window=1.5, peak_area_thres=1e5,
              min_scan=15, max_scan=200, max_peak=7):
    '''
    rt, ints from ms_chromatogram_list
    rt_window now set up for minutes
    '''

    # Important funciont, may need to be extracted out later
    def ms_chromatogram_list(mzml_scans, input_mz, error):
        '''
        Generate a peak list for specific input_mz
        over whole rt period from the mzml file
        ***Most useful function!
        '''
        retention_time = []
        intensity = []
        for scan in mzml_scans:
            # print(i)
            retention_time.append(scan.scan_time[0])

            target_mz, target_index = mz_locator(scan.mz, input_mz, error)
            if target_index == 'NA':
                intensity.append(0)
            else:
                intensity.append(sum(scan.i[target_index]))

        return retention_time, intensity

    rt, intensity = ms_chromatogram_list(mzml_scans, input_mz, error)

    # Get rt_window corresponded scan number
    scan_window = int((rt_window / (rt[int(len(intensity) / 2)] -
                                    rt[int(len(intensity) / 2) - 1])) / 2)

    # Get peak index
    indexes = peakutils.indexes(intensity, thres=thr, min_dist=min_d)

    result_dict = {}

    for index in indexes:
        h_range = index
        l_range = index
        base_intensity = peak_base

        # Get the higher and lower boundary
        while intensity[h_range] >= base_intensity:
            h_range += 1
            if h_range > len(intensity)-2:
                break
        while intensity[l_range] >= base_intensity:
            l_range -= 1
        # Output a range from the peak list

        peak_range = []
        if h_range - l_range >= min_scan:
            if rt[h_range] - rt[l_range] <= rt_window:
                peak_range = intensity[l_range:h_range]
            else:
                l_range = index - scan_window
                h_range = index + scan_window
                peak_range = intensity[l_range:h_range]
                # print(index + scan_window)

        # Intergration based on the simps function
        if len(peak_range) >= min_scan:
            integration_result = simps(peak_range)
            if integration_result >= peak_area_thres:
                if len(result_dict) == 0:
                    result_dict.update({index: [l_range, h_range,
                                                integration_result]})
                elif integration_result != list(result_dict.values())[-1][2]:
                    # Compare with previous item
                    # commented out s_window cus it wasn't used!!!
                    # s_window = abs(index - list(result_dict.keys())[-1])
                    # if s_window > min_scan_window:
                    result_dict.update({index: [l_range, h_range,
                                                integration_result]})

        # Filtering:
        # 1. delete results that l_range/h_range within 5 scans
        # 3. If still >5 then select top 5 results
        # list(result_dict.values())[-1]

    # Noise filter
    if len(result_dict) > max_peak:
        result_dict = {}

    return result_dict


def integration_plot(mzml_scans, input_mz, error,
                     peak_base=0.005, thr=0.02, min_d=1,
                     rt_window=2, peak_area_thres=1e5):

    result_dict = peak_pick(mzml_scans, input_mz, error)

    def ms_chromatogram_list(mzml_scans, input_mz, error):
        '''
        Generate a peak list for specific input_mz
        over whole rt period from the mzml file
        ***Most useful function!
        '''
        retention_time = []
        intensity = []
        for scan in mzml_scans:
            # print(i)
            retention_time.append(scan.scan_time[0])

            target_mz, target_index = mz_locator(scan.mz, input_mz, error)
            if target_index == 'NA':
                intensity.append(0)
            else:
                intensity.append(sum(scan.i[target_index]))

        return retention_time, intensity

    rt, ints = ms_chromatogram_list(mzml_scans, input_mz, error)

    plt.figure(figsize=(20, 10))
    plt.plot(rt, ints)
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    plt.xlabel('Retention Time(min)')
    plt.ylabel('Intensity')
    plt.title('Integration result')
    plt.xlim(0, rt[-1])
    plt.ylim(0, )

    for index in result_dict:
        print(('Peak retention time: {:0.2f} minute, Peak area: {: 0.1f}'
               .format(rt[index], result_dict[index][2])))
        plt.fill_between(rt[result_dict[index][0]: result_dict[index][1]],
                         ints[result_dict[index][0]: result_dict[index][1]])

    return



import pandas as pd
import urllib
import urllib.request
import json
import numpy as np
import matplotlib.pyplot as plt
import plotly.graph_objects as go
import webbrowser
import pyperclip


def massbank(frag_list, minsimilarity=500):
    url = 'https://mona.fiehnlab.ucdavis.edu/rest/similarity/search'
    values = {"spectrum": frag_list, "minSimilarity": minsimilarity}
    data_str = json.dumps(values)
    data = data_str.encode()
    # print(data)
    headers = {
      "content-type": "application/json",
    }
    req = urllib.request.Request(url, data, headers)
    response = urllib.request.urlopen(req)
    json_result = json.loads(response.read())
    result_list = []
    for result in json_result:
        name = result['hit']['compound'][0]['names'][0]['name']
        cpd_meta = result['hit']['compound'][0]['metaData']
        try:
            formula = [i for i in cpd_meta if i['name'] == 'molecular formula'
                       and i['category'] == 'none'][0]['value']
        except Exception:
            formula = "NA"
        try:
            inchi = result['hit']['compound'][0]['inchiKey']
        except Exception:
            inchi = "NA"
        ms_meta = result['hit']['metaData']
        try:
            instrument = [i for i in ms_meta if
                          i['name'] == 'instrument type'][0]['value']
        except Exception:
            instrument = "NA"
        mode = [i for i in ms_meta if i['name'] ==
                'ionization mode'][0]['value']
        try:
            energy = [i for i in ms_meta if i['name'] ==
                      'collision energy'][0]['value']
        except Exception:
            energy = "NA"
        try:
            p_type = [i for i in ms_meta if i['name'] ==
                      'precursor type'][0]['value']
        except Exception:
            p_type = "NA"
        score = result['score']
        link = ('https://mona.fiehnlab.ucdavis.edu/spectra/display/'
                + result['hit']['id'])
        result_list.append([name, formula, inchi, instrument,
                           energy, mode, p_type, score, link])
    d_result = pd.DataFrame(result_list,
                            columns=['name', 'formula', 'inchiKey',
                                     'instrument', 'CE', 'ionization mode',
                                     'precursor type', 'score', 'url'])

    def make_clickable(val):
        # target _blank to open new window
        return '<a target="_blank" href="{}">{}</a>'.format(val, val)

    d_result = d_result.style.format({'url': make_clickable})
    return d_result


def frag_massbank(mzml_scans, precursor, error=20, noise_thr=50, scan_index=0):
    p_range_l = precursor * (1 - error * 1e-6)
    p_range_h = precursor * (1 + error * 1e-6)
    frag_scan = []
    for scan in mzml_scans:
        if scan.ms_level == 2:
            precursor = scan.selected_precursors[0]['mz']
            p_intensity = scan.selected_precursors[0]['i']
            if precursor < p_range_h and precursor > p_range_l:
                frag_scan.append([precursor, p_intensity, scan])
    frag_scan.sort(key=lambda x: x[1], reverse=True)
    if len(frag_scan) != 0:
        print('Now showing index', scan_index, 'of',
              str(len(frag_scan)), 'total found scans')
        plot_scan = frag_scan[scan_index][2]
        drop_index = np.argwhere(plot_scan.i <= noise_thr)
        plot_scan.i = np.delete(plot_scan.i, drop_index)
        plot_scan.mz = np.delete(plot_scan.mz, drop_index)
        mz = plot_scan.mz
        ints = plot_scan.i

    for i in range(len(mz)):
        if i == 0:
            list_string = (str(round(mz[i], 4)) + ':' +
                           str(round(ints[i], 1)) + ' ')
        else:
            list_string += (str(round(mz[i], 4)) + ':' +
                            str(round(ints[i], 1)) + ' ')

    d_result = massbank(list_string)

    return d_result


def frag_comp(mzml_scans, precursor, error=20, scan_index=0, mona_index=0,
              noise_thr=50, interactive=False, search=False, source='MoNA'):
    '''
    Interactive spectrum plot with nearest retention time from the given scan
    mzml_scans: mzfile
    time: selected time for the scan
    '''
    p_range_l = precursor * (1 - error * 1e-6)
    p_range_h = precursor * (1 + error * 1e-6)
    frag_scan = []
    for scan in mzml_scans:
        if scan.ms_level == 2:
            precursor = scan.selected_precursors[0]['mz']
            p_intensity = scan.selected_precursors[0]['i']
            if precursor < p_range_h and precursor > p_range_l:
                frag_scan.append([precursor, p_intensity, scan])
    frag_scan.sort(key=lambda x: x[1], reverse=True)
    if len(frag_scan) != 0:
        print('Now showing index', scan_index, 'of',
              str(len(frag_scan)), 'total found scans')
        plot_scan = frag_scan[scan_index][2]
        drop_index = np.argwhere(plot_scan.i <= noise_thr)
        plot_scan.i = np.delete(plot_scan.i, drop_index)
        plot_scan.mz = np.delete(plot_scan.mz, drop_index)
        mz = plot_scan.mz
        ints = plot_scan.i
        rt = plot_scan.scan_time[0]
        print('Precursor:', round(plot_scan.selected_precursors[0]['mz'], 4),
              'precursor intensity:',
              round(plot_scan.selected_precursors[0]['i'], 1))
        print('Scan time:', round(plot_scan.scan_time[0], 2), 'minute')
        for i in range(len(mz)):
            if i == 0:
                list_string = (str(round(mz[i], 4)) + ':' +
                               str(round(ints[i], 1)) + ' ')
            else:
                list_string += (str(round(mz[i], 4)) + ':' +
                                str(round(ints[i], 1)) + ' ')
        url = 'https://mona.fiehnlab.ucdavis.edu/rest/similarity/search'
        values = {"spectrum": list_string, "minSimilarity": 500}

        data_str = json.dumps(values)
        data = data_str.encode()
        headers = {
          "content-type": "application/json",
        }
        req = urllib.request.Request(url, data, headers)
        response = urllib.request.urlopen(req)
        json_result = json.loads(response.read())

        mona_list = json_result[mona_index]['hit']['spectrum']
        spec1 = mona_list.split(' ')
        spec2 = [i.split(':') for i in spec1]
        mona_mz = [float(a) for a, b in spec2]
        mona_i = [-float(b) for a, b in spec2]

        # normalize scan intensity
        ints = ints/ints.max()*100

        if interactive is True:
            plt.clf()
            fig = go.Figure([go.Bar(x=mz, y=ints, marker_color='red',
                            width=0.5, hovertemplate='Int: %{y}' +
                            '<br>m/z: %{x}<br>')])
            fig.update_traces(marker_color='rgb(158,202,225)',
                              marker_line_color='rgb(0,0,0)',
                              marker_line_width=0.5, opacity=1)
            fig.update_layout(
                    title_text=(str(round(rt, 3)) +
                                ' min MS1 spectrum, input ' +
                                str(rt) + ' min'),
                    template='simple_white',
                    width=1000,
                    height=600,
                    xaxis={'title': 'm/z ratio'},
                    yaxis=dict(
                        showexponent='all',
                        exponentformat='e',
                        title='Intensity'))
            fig.show()

        elif interactive is False:
            fig, ax = plt.subplots(figsize=(12, 9))
            ax.bar(mz, ints, width=1.0, label='scan spectrum')
            ax.bar(mona_mz, mona_i, width=1.0, label='massbank spectrum')
            ax.axhline(y=0, color='k')
            plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
            ticks = ax.get_yticks()
            ax.set_yticklabels([int(abs(tick)) for tick in ticks])
            plt.xlabel('m/z')
            plt.ylabel('Relative Intensity %')
            plt.title('fragment spectrum')
            plt.legend()
            plt.xlim(0,)

        if search is True:
            for i in range(len(mz)):
                if i == 0:
                    list_string = (str(round(mz[i], 4)) + ' ' +
                                   str(round(ints[i], 1)) + '\r')
                else:
                    list_string += (str(round(mz[i], 4)) + ' ' +
                                    str(round(ints[i], 1)) + '\r')
            pyperclip.copy(list_string)
            if source == 'MoNA':
                (webbrowser.open(
                 "https://mona.fiehnlab.ucdavis.edu/spectra/search"))
            elif source == 'metfrag':
                webbrowser.open("https://msbi.ipb-halle.de/MetFragBeta/")
    else:
        print('No MS2 spectrum found!')
    return



from mss import visreader
from mss import mssmain
from mss import alignment
from mss import frag
from mss import mssdata
from mss import dm
from .version import __version__

__all__ = [visreader, mssmain, alignment,
           frag, mssdata, dm, __version__]



import unittest
import os
import pandas as pd
import numpy as np
from mss import dm
import mss
import sklearn
import example_data
test_path = os.path.join(mss.__path__[0], 'tests')
data_path = os.path.join(test_path, 'data')
file = 'sample1114.csv'
file_path = os.path.join(data_path, file)
test_path2 = os.path.join(example_data.__path__[0], 'clustering')
file_path2 = os.path.join(test_path2, file)
file2 = 'sample0815.csv'
file_path3 = os.path.join(test_path2, file2)
# Test update: how to write this type of test in chain
# output from one function is input for next function


class test_dm(unittest.TestCase):
    def test_data_prep(self):
        '''test data prep function'''
        d_ms = pd.read_csv(file_path)
        keys = ['CEC', 'Blank', 'ISTD', 'Wash', 'Shutdown']
        d_test = dm.data_prep(d_ms, keys, rt_range=[1, 30],
                              mz_range=[200, 800], area_thres=500,
                              simp_summary=False)
        assert len(d_test) == 5299, 'dataframe length error!'
        # More assert here
        return

    def test_ms_cluster(self):
        keys = ['CEC', 'Blank', 'ISTD', 'Wash', 'Shutdown']
        d_ms = pd.read_csv(file_path)
        d_sample = dm.data_prep(d_ms, keys, rt_range=[1, 30],
                                mz_range=[200, 800], area_thres=500,
                                simp_summary=False)
        d_test = dm.ms_cluster(d_sample, ['SR520-Cal'], 'linear',
                               d_reduce=False, visual=False,
                               cluster_method='optics',
                               eps=0.6, min_samples=10)
        assert sum(d_test['label'] == np.nan) == 0, 'invalid label presents'
        assert len(list(set(d_test['label']))) == 16, 'incorrect clustering'
        assert d_test.shape[1]-d_sample.shape[1] == 1, 'column insert error'
        return

    def test_source_label(self):
        keys = ['CEC', 'Blank', 'ISTD', 'Wash', 'Shutdown']
        d_ms = pd.read_csv(file_path)
        d_sample = dm.data_prep(d_ms, keys, rt_range=[1, 30],
                                mz_range=[200, 800], area_thres=500,
                                simp_summary=False)
        sourcelist = ['Coulter', 'Crescent', 'Miller',
                      'Swan', 'SR520-Cal-in-DI_1000mL']
        # Needs adjustment
        d_test = dm.source_label(d_sample, sourcelist,
                                 area_thres=50000, concat=True)
        assert d_test.shape[1] - d_sample.shape[1] == 11, 'column number error'
        assert d_test['source'].dtype == object, 'source column type error'
        return

    def test_source_report(self):
        keys = ['CEC', 'Blank', 'ISTD', 'Wash', 'Shutdown']
        d_ms = pd.read_csv(file_path)
        d_sample = dm.data_prep(d_ms, keys, rt_range=[1, 30],
                                mz_range=[200, 800], area_thres=500,
                                simp_summary=False)
        sourcelist = ['Coulter', 'Crescent', 'Miller', 'Swan',
                      'SR520-Cal-in-DI_1000mL']
        # Needs adjustment
        d_label = dm.source_label(d_sample, sourcelist, area_thres=50000,
                                  concat=True)
        d_test = dm.source_report(d_label, ['Coulter', 'Crescent',
                                  'Miller', 'Swan', 'SR520-Cal-in-DI_1000mL'],
                                  ['Mix'], method='multiple',
                                  pa_thres=10000, CV_thres=2)
        assert d_test.shape == (27, 16), 'dataframe shape error'
        assert d_test['sample'].dtype == object, 'sample column type error'
        return

    def test_trend_calc(self):
        keys = ['CEC', 'Blank', 'ISTD', 'Wash', 'Shutdown']
        key = ['SR520-Cal']
        d_ms = pd.read_csv(file_path)
        d_sample = dm.data_prep(d_ms, keys, rt_range=[1, 30],
                                mz_range=[200, 800], area_thres=500,
                                simp_summary=False)
        d_test = dm.trend_calc(d_sample, key, min_size=5,
                               normalization='zscore',
                               method='pearsonr', visual=False)
        d_test1 = dm.trend_calc(d_sample, key, min_size=5,
                                normalization='log',
                                method='pearsonr', visual=False)
        d_test2 = dm.trend_calc(d_sample, key, min_size=5,
                                normalization='linear',
                                method='pearsonr', visual=False)
        assert len(d_test) > 0, 'no output'
        assert type(d_test) == pd.core.frame.DataFrame, 'wrong output'
        assert len(d_test1) > 0, 'no output'
        assert type(d_test1) == pd.core.frame.DataFrame, 'wrong output'
        assert len(d_test2) > 0, 'no output'
        assert type(d_test2) == pd.core.frame.DataFrame, 'wrong output'
        return

    def test_batch_alignment(self):
        d_model = pd.read_csv(file_path2)
        d_test = pd.read_csv(file_path3)
        d_merge = dm.batch_alignment(d_model, d_test)
        assert len(d_merge) > 0, 'alignment went wrong'
        assert type(d_merge) == pd.core.frame.DataFrame, 'wrong output'
        return

    def test_transpose(self):
        keys = ['CEC', 'Blank', 'ISTD', 'Wash', 'Shutdown']
        d_ms = pd.read_csv(file_path)
        d_sample = dm.data_prep(d_ms, keys, rt_range=[1, 30],
                                mz_range=[200, 800], area_thres=500,
                                simp_summary=False)
        d_sample2 = dm.ms_cluster(d_sample, ['SR520-Cal'], 'linear',
                                  d_reduce=False, visual=False,
                                  eps=0.6, min_samples=10)
        d_ref = pd.read_csv(file_path3)
        d_model = d_sample2[d_sample2['label'] != -1]
        d_merge = dm.batch_alignment(d_model, d_ref)
        dilu_col = ['SR520-Cal', 'SR520_Cal']
        d_transpose = dm.transpose(d_merge, dilu_col)
        assert len(d_transpose) > 0, 'transpose went wrong'
        assert type(d_transpose) == pd.core.frame.DataFrame, 'wrong output'
        return

    def test_feature_model(self):
        keys = ['CEC', 'Blank', 'ISTD', 'Wash', 'Shutdown']
        d_ms = pd.read_csv(file_path)
        d_sample = dm.data_prep(d_ms, keys, rt_range=[1, 30],
                                mz_range=[200, 800], area_thres=500,
                                simp_summary=False)
        d_sample2 = dm.ms_cluster(d_sample, ['SR520-Cal'], 'linear',
                                  d_reduce=False, visual=False,
                                  eps=0.6, min_samples=10)
        d_ref = pd.read_csv(file_path3)
        d_model = d_sample2[d_sample2['label'] != -1]
        d_merge = dm.batch_alignment(d_model, d_ref)
        dilu_col = ['SR520-Cal', 'SR520_Cal']
        d_transpose = dm.transpose(d_merge, dilu_col)
        reg = dm.feature_model(d_transpose, cluster_algorithm=False,
                               model_method='linear_reg', report=False)
        reg1 = dm.feature_model(d_transpose, cluster_algorithm=True,
                                model_method='ensemble_rf', report=False)
        reg2 = dm.feature_model(d_transpose, cluster_algorithm=False,
                                model_method='ensemble_bagging_dt',
                                report=False)
        reg3 = dm.feature_model(d_transpose, cluster_algorithm=True,
                                model_method='ensemble_bagging_svc',
                                report=False)
        assert type(reg) == sklearn.linear_model._base.LinearRegression, \
            'wrong regression applied'
        assert type(reg1) == list, 'wrong regression applied'
        assert type(reg2) == sklearn.ensemble._bagging.BaggingRegressor, \
            'wrong regression applied'
        assert type(reg3) == list, 'wrong regression applied'
        return

    def test_cluster_pred(self):
        keys = ['CEC', 'Blank', 'ISTD', 'Wash', 'Shutdown']
        d_ms = pd.read_csv(file_path)
        d_sample = dm.data_prep(d_ms, keys, rt_range=[1, 30],
                                mz_range=[200, 800], area_thres=500,
                                simp_summary=False)
        d_sample2 = dm.ms_cluster(d_sample, ['SR520-Cal'], 'linear',
                                  d_reduce=False, visual=False,
                                  eps=0.6, min_samples=10)
        d_ref = pd.read_csv(file_path3)
        d_model = d_sample2[d_sample2['label'] != -1]
        d_merge = dm.batch_alignment(d_model, d_ref)
        dilu_col = ['SR520-Cal', 'SR520_Cal']
        d_transpose = dm.transpose(d_merge, dilu_col)
        reg = dm.feature_model(d_transpose, cluster_algorithm=True,
                               model_method='ensemble_bagging_dt',
                               report=False)
        reg1 = dm.feature_model(d_transpose, cluster_algorithm=True,
                                model_method='ensemble_rf', report=False)
        dilu_test = dm.cluster_pred(reg, d_merge, '20181114_CoulterCreek_1',
                                    selected_cluster=[0, 1, 3, 5, 7, 9],
                                    model_merge='weighted_size')
        dilu_test2 = dm.cluster_pred(reg1, d_merge, '20181114_CoulterCreek_1',
                                     selected_cluster=[0, 1, 3, 5, 7, 9],
                                     model_merge='average')
        assert dilu_test >= 0, 'Invalid dilution prediction'
        assert dilu_test2 >= 0, 'Invalid dilution prediction'
        return



# import os
# import sys
import setuptools

# Get appropriate version and release info
# ver_file = os.path.join('mss', 'version.py')

# Grabs description from the README file
# with open(ver_file) as f:
#     exec(f.read())

# Give setuptools a hit to complain if it's too old of a version
# 24.2.0 added the python_requires option
# Enables setuptools to install wheel on-the-fly
#SETUP_REQUIRES = ['setuptools >= 24.2.0']
#SETUP_REQUIRES += ['wheel'] if 'bdist_wheel' in sys.argv else []

setuptools.setup(name = 'mass-suite',
	  version = '1.0.0',
      description = 'A package for HRMS data analysis',
      long_description = open('README.md', 'r').read(),
      long_description_content_type = 'text/markdown',
      url = 'https://github.com/XiminHu/mass-suite',
      author = 'Ximin Hu, Derek Mar, Nozomi Suzuki, Bowei Zhang',
      author_email = 'xhu66@uw.edu',
      packages = setuptools.find_packages(),
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
        python_requires='>=3.6',
)



import os
import unittest
import mss
from mss import mssmain

test_path = os.path.join(mss.__path__[0], 'tests')
data_path = os.path.join(test_path, 'data')
file = 'ex_1.mzML'
file_path = os.path.join(data_path, file)


class test_mssmain(unittest.TestCase):

    def test_get_scans(self):
        """Tests stack functionlity"""
        test_scan = mssmain.get_scans(file_path)
        assert type(test_scan) == list, "output type is wrong"
        assert len(test_scan) == 4061, "output length is wrong"
        assert len(test_scan[0].mz) != 0, "mz array didn't read in"
        assert len(test_scan[0].i) != 0, "i array didn't read in"
        return

    def test_noise_removal(self):
        test_scan = mssmain.get_scans(file_path)
        mssmain.noise_removal(test_scan)
        assert sum(test_scan[0].i <= 5000) == 0,\
            "noise didn't properly removed"
        return

    def test_mzlocator(self):
        test_scan = mssmain.get_scans(file_path)
        mzlist = mssmain.mz_locator(test_scan[0].mz, 117.113, 20)
        assert type(mzlist) == tuple, "output type is wrong"
        assert mzlist[0][0] - 117.113 <= 20 * 117.113,\
            "selected mz is above defined error range"
        assert mzlist[1][0] != 0, "index error"
        return

    def test_peak_pick(self):  # More to add during later development
        test_scan = mssmain.get_scans(file_path)
        test_dict = mssmain.peak_pick(test_scan, 299.146, 20)
        assert type(test_dict) == dict, "output type is wrong"
        assert len(test_dict.keys()) != 0, "didn't find the peak"
        assert list(test_dict.values())[0][2] - 136350 <= 1,\
            "intergration result error"
        return

    def test_peak_list(self):  # More to add during later development
        test_scan = mssmain.get_scans(file_path)
        d_test = mssmain.peak_list(test_scan[:200], 20, enable_score=True)
        assert d_test.shape[1] == 5, "score column is off"
        assert d_test.score.dtype == 'int64', "wrong score type"
        return

#    def test_batch_scans(self):
#       return

#    def test_batch_peak(self):
#        return
